#include <math.h>

double digamma(double x) {
  double r, f1, f2, f3, f4;
  r = 0;

  while (x < 5) {
    r -= 1 / x;
    x += 1;
  }

  f1 = 1 / (x * x);
  f2 = f1 * f1;
  f3 = f2 * f1;
  f4 = f2 * f2;

  static const double c1 = -1 / 12.0;
  static const double c2 = 1 / 120.0;
  static const double c3 = -1 / 252.0;
  static const double c4 = 1 / 240.0;
  static const double c5 = -1 / 132.0;
  static const double c6 = 691 / 32760.0;
  static const double c7 = -1 / 12.0;
  static const double c8 = 3617 / 8160.0;

  double A = c1 + c2 * f1 + c3 * f2 + c4 * f3;
  double B = c5 * f1 + c6 * f2 + c7 * f3 + c8 * f4;
  return r + log(x) - 0.5 / x + f1 * A + f4 * B;
}

#define USE_RINTERNALS

#include <R.h>
#include <Rinternals.h>

void set_rownames(SEXP x);
void set_names(SEXP x);

SEXP mat2vec(SEXP x) {
  char type = TYPEOF(x);
  if (!isMatrix(x))
    error("'x' must be a matrix");
  int nRow = INTEGER(getAttrib(x, R_DimSymbol))[0];
  int nCol = INTEGER(getAttrib(x, R_DimSymbol))[1];
  SEXP output = PROTECT(allocVector(VECSXP, nCol));

#define HANDLE_CASE(ACCESSOR, CTYPE)                       \
  {                                                        \
    for (int i = 0; i < nCol; ++i) {                       \
      SET_VECTOR_ELT(output, i, allocVector(type, nRow));  \
      CTYPE* output_ptr = ACCESSOR(VECTOR_ELT(output, i)); \
      CTYPE* x_ptr = ACCESSOR(x);                          \
      for (int j = 0; j < nRow; ++j) {                     \
        output_ptr[j] = x_ptr[nRow * i + j];               \
      }                                                    \
    }                                                      \
    break;                                                 \
  }

#define HANDLE_CASE_MEMCPY(ACCESSOR, CTYPE)                                  \
  {                                                                          \
    for (int i = 0; i < nCol; ++i) {                                         \
      char sz = sizeof(CTYPE);                                               \
      SET_VECTOR_ELT(output, i, allocVector(type, nRow));                    \
      SEXP elt = VECTOR_ELT(output, i);                                      \
      memcpy(                                                                \
          (char*)DATAPTR(elt), (char*)DATAPTR(x) + (i* nRow* sz), nRow* sz); \
    }                                                                        \
    break;                                                                   \
  }

  switch (type) {
    case INTSXP:
      HANDLE_CASE_MEMCPY(INTEGER, int);
    case REALSXP:
      HANDLE_CASE_MEMCPY(REAL, double);
    case LGLSXP:
      HANDLE_CASE_MEMCPY(INTEGER, int);
    case STRSXP:
      HANDLE_CASE(STRING_PTR, SEXP);
    default:
      error("Unhandled SEXP type '%s'", type2char(type));
  }

#undef HANDLE_CASE
#undef HANDLE_CASE_MEMCPY

  setAttrib(
      output, R_NamesSymbol, VECTOR_ELT(getAttrib(x, R_DimNamesSymbol), 1));

  UNPROTECT(1);
  return output;
}

#undef USE_RINTERNALS

#define USE_RINTERNALS

#include <R.h>
#include <Rinternals.h>
#include <stdbool.h>

// utils.c
char max_type(SEXP, SEXP);

// a function that operates like R's 'rep(..., each=each)',
// but only works for characters
SEXP rep_each_char(SEXP x, SEXP id_ind_, int each) {

  SEXP out;
  int* id_ind = INTEGER(id_ind_);
  int len = length(id_ind_);
  PROTECT(out = allocVector(STRSXP, len * each));
  int counter = 0;
  SEXP* ptr = STRING_PTR(x);
  SEXP* out_ptr = STRING_PTR(out);
  for (int i = 0; i < len; ++i) {
    for (int j = 0; j < each; ++j) {
      out_ptr[counter] = ptr[id_ind[i]];
      ++counter;
    }
  }
  UNPROTECT(1);
  return out;
}

// some macro definitions used for repeating a vector
// n times

// simply using memcpy with strings is dangerous
#define HANDLE_CASE_STRING                           \
  case STRSXP: {                                     \
    int counter = 0;                                 \
    PROTECT(out = allocVector(STRSXP, len * times)); \
    SEXP* ptr = STRING_PTR(x);                       \
    SEXP* out_ptr = STRING_PTR(out);                 \
    for (int i = 0; i < times; ++i) {                \
      for (int j = 0; j < len; ++j) {                \
        out_ptr[counter] = ptr[j];                   \
        ++counter;                                   \
      }                                              \
    }                                                \
    UNPROTECT(1);                                    \
    return out;                                      \
  }

#define HANDLE_CASE(RTYPE, CTYPE)                                             \
  case RTYPE: {                                                               \
    char sz = sizeof(CTYPE);                                                  \
    PROTECT(out = allocVector(RTYPE, len * times));                           \
    for (int i = 0; i < times; ++i) {                                         \
      memcpy((char*)DATAPTR(out) + (i* len* sz), (char*)DATAPTR(x), len* sz); \
    }                                                                         \
    UNPROTECT(1);                                                             \
    return out;                                                               \
  }

SEXP stack_vector(SEXP x, int times) {
  SEXP out;
  int len = length(x);
  switch (TYPEOF(x)) {
    HANDLE_CASE(INTSXP, int);
    HANDLE_CASE(REALSXP, double);
    HANDLE_CASE(LGLSXP, int);
    HANDLE_CASE_STRING;
  }

  // if we've reached here, we have an unhandled / incompatible SEXP type
  error("argument is of incompatible type '%s'", type2char(TYPEOF(x)));
  return R_NilValue;
}

#undef HANDLE_CASE

// checks if all values in a VECSXP x are of the same type
bool diff_types(SEXP x, SEXP val_ind_) {
  if (TYPEOF(x) != VECSXP) {
    error("Expected a VECSXP but got a '%s'", type2char(TYPEOF(x)));
  }
  int n = length(val_ind_);
  int* val_ind = INTEGER(val_ind_);
  char type = TYPEOF(VECTOR_ELT(x, val_ind[0]));
  for (int i = 1; i < n; ++i) {
    if (TYPEOF(VECTOR_ELT(x, val_ind[i])) != type) {
      return true;
    }
  }
  return false;
}

// [[register]]
SEXP melt_dataframe(SEXP x,
                    SEXP id_ind_,
                    SEXP val_ind_,
                    SEXP variable_name,
                    SEXP value_name) {

  if (length(x) == 0) {
    error("Can't melt a data.frame with 0 columns");
  }

  if (length(VECTOR_ELT(x, 0)) == 0) {
    error("Can't melt a data.frame with 0 rows");
  }

  int* id_ind = INTEGER(id_ind_);
  int* val_ind = INTEGER(val_ind_);

  int nColStack = length(id_ind_);
  int nColRep = length(val_ind_);

  int nRow = length(VECTOR_ELT(x, 0));
  int out_nRow = nRow * nColRep;
  int out_nCol = nColStack + 2;

  char mt = max_type(x, val_ind_);
  if (mt > STRSXP) {
    error("Error: cannot melt data.frames w/ elements of type '%s'",
          CHAR(type2str(mt)));
  }

  if (diff_types(x, val_ind_)) {
    warning("Coercing type of 'value' variables to '%s'", CHAR(type2str(mt)));
  }

  SEXP out;
  PROTECT(out = allocVector(VECSXP, out_nCol));

  // populate the value array
  SEXP value_SEXP;

#define HANDLE_CASE(RTYPE, CTYPE)                                    \
  case RTYPE: {                                                      \
    PROTECT(value_SEXP = allocVector(RTYPE, value_len));             \
    SEXP tmp;                                                        \
    for (int i = 0; i < nColRep; ++i) {                              \
      if (TYPEOF(VECTOR_ELT(x, val_ind[i])) != mt) {                 \
        tmp = PROTECT(coerceVector(VECTOR_ELT(x, val_ind[i]), mt));  \
      } else {                                                       \
        tmp = VECTOR_ELT(x, val_ind[i]);                             \
      }                                                              \
      memcpy((char*)DATAPTR(value_SEXP) + (i* nRow * sizeof(CTYPE)), \
             (char*)DATAPTR(tmp),                                    \
             nRow * sizeof(CTYPE));                                  \
      if (TYPEOF(VECTOR_ELT(x, val_ind[i])) != mt) {                 \
        UNPROTECT(1);                                                \
      }                                                              \
    }                                                                \
    break;                                                           \
  }

  int value_len = nColRep * nRow;
  int value_type = mt;
  switch (value_type) {
    HANDLE_CASE(INTSXP, int);
    HANDLE_CASE(REALSXP, double);
    HANDLE_CASE(LGLSXP, int);
    case STRSXP: {
      int counter = 0;
      SEXP* curr_str_vec_ptr;
      SEXP tmp;
      PROTECT(value_SEXP = allocVector(STRSXP, value_len));
      for (int i = 0; i < nColRep; ++i) {
#define curr_str_vec (VECTOR_ELT(x, val_ind[i]))
        if (TYPEOF(curr_str_vec) != STRSXP) {
          if (isFactor(curr_str_vec)) {
            PROTECT(tmp = asCharacterFactor(curr_str_vec));
          } else {
            PROTECT(tmp = coerceVector(curr_str_vec, STRSXP));
          }
          curr_str_vec_ptr = STRING_PTR(tmp);
        } else {
          curr_str_vec_ptr = STRING_PTR(curr_str_vec);
        }
#undef curr_str_vec
        SEXP* value_SEXP_ptr = STRING_PTR(value_SEXP);
        for (int j = 0; j < nRow; ++j) {
          value_SEXP_ptr[counter] = curr_str_vec_ptr[j];
          ++counter;
        }
        if (TYPEOF(VECTOR_ELT(x, val_ind[i])) != mt) {
          UNPROTECT(1);
        }
      }
      break;
    }
    default:
      error("Unsupported RTYPE encountered");
  }

#undef HANDLE_CASE

  // generate the id variables, and assign them on generation
  // we need to convert factors if necessary
  for (int i = 0; i < nColStack; ++i) {
    SET_VECTOR_ELT(out, i, stack_vector(VECTOR_ELT(x, id_ind[i]), nColRep));
    if (isFactor(VECTOR_ELT(x, id_ind[i]))) {
      setAttrib(VECTOR_ELT(out, i), R_ClassSymbol, mkString("factor"));
      setAttrib(VECTOR_ELT(out, i),
                R_LevelsSymbol,
                getAttrib(VECTOR_ELT(x, id_ind[i]), R_LevelsSymbol));
    }
  }

  // assign the names, values
  SET_VECTOR_ELT(out,
                 nColStack,
                 rep_each_char(getAttrib(x, R_NamesSymbol), val_ind_, nRow));
  SET_VECTOR_ELT(out, nColStack + 1, value_SEXP);
  UNPROTECT(1);  // value_SEXP

  // set the row names
  SEXP row_names;
  PROTECT(row_names = allocVector(INTSXP, out_nRow));
  int* row_names_ptr = INTEGER(row_names);
  for (int i = 0; i < out_nRow; ++i) {
    row_names_ptr[i] = i + 1;
  }
  setAttrib(out, R_RowNamesSymbol, row_names);
  UNPROTECT(1);  // row_names

  // set the class to data.frame
  setAttrib(out, R_ClassSymbol, mkString("data.frame"));

  // set the names
  SEXP names = getAttrib(x, R_NamesSymbol);
  SEXP names_out;
  PROTECT(names_out = allocVector(STRSXP, out_nCol));

  SEXP* names_ptr = STRING_PTR(names);
  SEXP* names_out_ptr = STRING_PTR(names_out);
  for (int i = 0; i < nColStack; ++i) {
    names_out_ptr[i] = names_ptr[id_ind[i]];
  }

  SET_STRING_ELT(names_out, nColStack, STRING_ELT(variable_name, 0));
  SET_STRING_ELT(names_out, nColStack + 1, STRING_ELT(value_name, 0));
  setAttrib(out, R_NamesSymbol, names_out);
  UNPROTECT(1);  // names_out

  UNPROTECT(1);  // out
  return out;
}

#undef USE_RINTERNALS

#define USE_RINTERNALS

#include <R.h>
#include <Rinternals.h>

SEXP rep_row_names(SEXP x, int times) {
  SEXP out;
  int len = length(x);
  int counter = 0;
  PROTECT(out = allocVector(STRSXP, len * times));
  SEXP* x_ptr = STRING_PTR(x);
  SEXP* out_ptr = STRING_PTR(out);
  for (int i = 0; i < times; ++i) {
    for (int j = 0; j < len; ++j) {
      out_ptr[counter] = x_ptr[j];
      // SET_STRING_ELT( out, counter, STRING_ELT(x, j) );
      ++counter;
    }
  }
  UNPROTECT(1);
  return out;
}
SEXP rep_col_names(SEXP x, int each) {
  SEXP out;
  int len = length(x);
  PROTECT(out = allocVector(STRSXP, len * each));
  int counter = 0;
  SEXP* ptr = STRING_PTR(x);
  SEXP* out_ptr = STRING_PTR(out);
  for (int i = 0; i < len; ++i) {
    for (int j = 0; j < each; j++) {
      out_ptr[counter] = ptr[i];
      // SET_STRING_ELT( out, counter, ptr[i] );
      ++counter;
    }
  }
  UNPROTECT(1);
  return out;
}

SEXP matrix_to_vector(SEXP x, int size) {

  SEXP out;
  switch (TYPEOF(x)) {
    case INTSXP: {
      PROTECT(out = allocVector(INTSXP, size));
      int* mat_ptr = INTEGER(x);
      int* out_ptr = INTEGER(out);
      for (int i = 0; i < size; ++i) {
        out_ptr[i] = mat_ptr[i];
      }
      UNPROTECT(1);
      return out;
    }
    case REALSXP: {
      PROTECT(out = allocVector(REALSXP, size));
      double* mat_ptr = REAL(x);
      double* out_ptr = REAL(out);
      for (int i = 0; i < size; ++i) {
        out_ptr[i] = mat_ptr[i];
      }
      UNPROTECT(1);
      return out;
    }
    case LGLSXP: {
      PROTECT(out = allocVector(LGLSXP, size));
      int* mat_ptr = LOGICAL(x);
      int* out_ptr = LOGICAL(out);
      for (int i = 0; i < size; ++i) {
        out_ptr[i] = mat_ptr[i];
      }
      UNPROTECT(1);
      return out;
    }
    case STRSXP: {
      PROTECT(out = allocVector(STRSXP, size));
      SEXP* mat_ptr = STRING_PTR(x);
      SEXP* out_ptr = STRING_PTR(out);
      for (int i = 0; i < size; ++i) {
        out_ptr[i] = mat_ptr[i];
      }
      UNPROTECT(1);
      return out;
    }
    default: { return R_NilValue; }
  }
}

// [[register]]
SEXP melt_matrix(SEXP x) {

  SEXP row, col, out, out_row_names;

  int nRow = nrows(x);
  int nCol = ncols(x);
  int out_nRow = nRow * nCol;
  int out_nCol = 3;

  // the output will be a 3 column data.frame
  // 1: repeated row names / indexes
  // 2: repeated col names / indexes
  // 3: values

  SEXP row_names, col_names;
  const char* row_names_char;
  const char* col_names_char;
  GetMatrixDimnames(
      x, &row_names, &col_names, &row_names_char, &col_names_char);

  PROTECT(out = allocVector(VECSXP, 3));
  int counter;

  // row indices
  if (isNull(row_names)) {
    PROTECT(row = allocVector(INTSXP, out_nRow));
    int* row_ptr = INTEGER(row);
    counter = 0;
    for (int i = 0; i < nCol; ++i) {
      for (int j = 0; j < nRow; ++j) {
        row_ptr[counter] = j + 1;
        ++counter;
      }
    }
  } else {
    PROTECT(row = rep_row_names(row_names, nCol));
  }

  // col indices
  if (isNull(col_names)) {
    PROTECT(col = allocVector(INTSXP, out_nRow));
    int* col_ptr = INTEGER(col);
    counter = 0;
    for (int i = 0; i < nCol; ++i) {
      for (int j = 0; j < nRow; ++j) {
        col_ptr[counter] = i + 1;
        ++counter;
      }
    }
  } else {
    PROTECT(col = rep_col_names(col_names, nRow));
  }

  // dim_names is a list; 1st entry is row names, 2nd is col names
  SET_VECTOR_ELT(out, 0, row);
  SET_VECTOR_ELT(out, 1, col);
  SET_VECTOR_ELT(out, 2, matrix_to_vector(x, out_nRow));

  // set row names
  PROTECT(out_row_names = allocVector(INTSXP, out_nRow));
  int* row_names_ptr = INTEGER(out_row_names);
  for (int i = 0; i < out_nRow; ++i) {
    row_names_ptr[i] = i + 1;
  }
  setAttrib(out, R_RowNamesSymbol, out_row_names);
  UNPROTECT(1);

  // set class
  setAttrib(out, R_ClassSymbol, mkString("data.frame"));

  // set names
  SEXP names;
  PROTECT(names = allocVector(STRSXP, out_nCol));
  SET_STRING_ELT(names, 0, mkChar("row"));
  SET_STRING_ELT(names, 1, mkChar("col"));
  SET_STRING_ELT(names, 2, mkChar("value"));
  setAttrib(out, R_NamesSymbol, names);
  UNPROTECT(1);

  // unprotect the rest of the stuff from earlier
  UNPROTECT(3);
  return out;
}

#undef USE_RINTERNALS

#define USE_RINTERNALS

#include <R.h>
#include <Rinternals.h>

// [[register]]
SEXP transpose_list(SEXP x) {

  int n = length(x);
  int N = length(VECTOR_ELT(x, 0));
  int type = TYPEOF(VECTOR_ELT(x, 0));

  SEXP tmp;
  SEXP output = PROTECT(allocVector(VECSXP, N));

#define HANDLE_CASE(RTYPE, CTYPE, ACCESSOR)      \
  {                                              \
    tmp = PROTECT(allocVector(RTYPE, n));        \
    CTYPE* ptr = ACCESSOR(tmp);                  \
    for (int j = 0; j < N; ++j) {                \
      for (int i = 0; i < n; ++i) {              \
        ptr[i] = ACCESSOR(VECTOR_ELT(x, i))[j];  \
      }                                          \
      SET_VECTOR_ELT(output, j, duplicate(tmp)); \
    }                                            \
    UNPROTECT(1);                                \
    break;                                       \
  }

  switch (type) {
    case LGLSXP:
      HANDLE_CASE(LGLSXP, int, LOGICAL);
    case INTSXP:
      HANDLE_CASE(INTSXP, int, INTEGER);
    case REALSXP:
      HANDLE_CASE(REALSXP, double, REAL);
    case STRSXP:
      HANDLE_CASE(STRSXP, SEXP, STRING_PTR);
    default: {
      Rf_error("Can't transpose a list with elements of type '%s'",
               CHAR(type2str(type)));
    }
  }

  UNPROTECT(1);
  return output;
}

#undef USE_RINTERNALS

#define USE_RINTERNALS

#include <R.h>
#include <Rinternals.h>

char max_type1(SEXP x) {
  if (TYPEOF(x) != VECSXP) {
    error("Expected a VECSXP but got a '%s'", type2char(TYPEOF(x)));
  }
  int n = length(x);
  char max_type = -1;
  char tmp = -1;
  for (int i = 0; i < n; ++i) {
    // factors should mean we coerce to string
    if (isFactor(VECTOR_ELT(x, i))) {
      if (STRSXP > max_type) {
        max_type = STRSXP;
      }
    } else if ((tmp = TYPEOF(VECTOR_ELT(x, i))) > max_type) {
      max_type = tmp;
    }
  }
  return max_type;
}

char max_type(SEXP x, SEXP ind_) {
  if (TYPEOF(x) != VECSXP) {
    error("Expected a VECSXP but got a '%s'", type2char(TYPEOF(x)));
  }
  int n = length(ind_);
  int* ind = INTEGER(ind_);
  char max_type = -1;
  char tmp = -1;
  for (int i = 0; i < n; ++i) {
    // factors should mean we coerce to string
    if (isFactor(VECTOR_ELT(x, ind[i]))) {
      if (STRSXP > max_type) {
        max_type = STRSXP;
      }
    } else if ((tmp = TYPEOF(VECTOR_ELT(x, ind[i]))) > max_type) {
      max_type = tmp;
    }
  }
  return max_type;
}

#undef USE_RINTERNALS

